@@ -28,6 +28,8 @@ use Arch::Tree;
 use Arch::Changeset;
 use Arch::TempFiles;
 
+use vars qw(@WINDOWS);
+
 # exe title is_unique name label descr
 my @gui_infos = qw(
 	archway  ArchWay    1  main   Main       Main-Dispatcher
@@ -78,6 +80,33 @@ while (
 	};
 }
 
+# guess ssh askpass
+if (! exists $ENV{SSH_ASKPASS}) {
+	TRY:
+	foreach my $dir (split(/:/, $ENV{PATH}), '/usr/lib/ssh') {
+		foreach my $prog (qw(ssh-askpass x11-ssh-askpass)) {
+			my $full_prog = "$dir/$prog";
+		
+			if (-x $full_prog) {
+				$ENV{SSH_ASKPASS} = $full_prog;
+				last TRY;
+			}
+		}
+	}
+}
+
+# use archway provided ssh-askpass if possible
+if (! exists $ENV{SSH_ASKPASS}) {
+	foreach my $dir ($FindBin::Bin, '/usr/share/archway', '/usr/lib/archway') {
+		my $full_prog = "$dir/archway-ssh-askpass";
+		
+		if (-x $full_prog) {
+			$ENV{SSH_ASKPASS} = $full_prog;
+			last;
+		}
+	}
+}
+
 sub run ($;$) {
 	my $class = shift;
 	my $initial_gui_name = shift;
@@ -292,6 +321,8 @@ sub open_main_window ($%) {
 	$self->check_tool_checkboxes($name);
 
 	$window->signal_connect(destroy => sub {
+		@WINDOWS = grep { $_ ne $window } @WINDOWS;
+
 		$gui_status->{window} = undef if $is_unique;
 		@{$gui_status->{windows}} =
 			grep { $_ != $window } @{$gui_status->{windows}};
@@ -303,6 +334,9 @@ sub open_main_window ($%) {
 		Gtk2->main_quit() unless $self->{num_main_windows} > 0;
 	});
 	$window->show_all;
+
+	push @WINDOWS, $window;
+
 	return $window;
 }
 
@@ -444,4 +478,118 @@ sub gui_logo_file ($$) {
 	return $logo_file;
 }
 
+# This code is simply evil ... nuff said.
+
+# We want a reactive GUI while tla is working, so we fork off and run
+# tla asynchronously.  But lacking coroutines or continuations, we
+# have to nest Gtk2->main calls.  Which in itself isn't all that bad,
+# unless the GUI is still enabled during the nested Gtk2->main call
+# (which is exactly what we do here), because we might end up being
+# called again.
+
+# In this situation, we can't just call Gtk2->main_quit whenever a tla
+# process finishes, because it would probably terminate the wrong main
+# loop and we would end up with incomplete output.
+
+# Instead we have to take a little note (@AWRT_DONE) when one of the
+# outer processes finishes and only ever terminate the innermost loop
+# directly.  Still, we need to quit every main loop once.  So just
+# before returning from an inner run_tla call, we check if the
+# containing call has already finished and terminate its main loop as
+# apropriate.
+
+# KNOWN PROBLEMS:
+#   o bad interaction with quit action
+
+use Arch::Util;
+use Arch::Run;
+package Arch::Util;
+
+# detach from console if SSH_ASKPASS is set
+Arch::Run::set_detach_console(exists $ENV{SSH_ASKPASS});
+
+our @AWRT_DONE;
+our $LAST_PID;
+
+BEGIN { undef &run_tla; }
+
+sub run_tla (@) {
+	my $arg0 = shift || die "no arch subcommand\n";
+	my @cmd = (Arch::Util::arch_backend, split(/ /, $arg0), @_);
+
+	my $level = @AWRT_DONE;
+	push @AWRT_DONE, 0;
+
+	# run tla process
+	my ($data, $exit) = ('', undef);
+	my $pid = Arch::Run::run_async(
+		command => \@cmd,
+		mode    => Arch::Run::RAW(),
+		datacb  => sub { $data .= $_[0] },
+		exitcb  => sub {
+			$exit = $_[0];
+			$AWRT_DONE[$level] = 1;
+
+			# call Gtk2->main_quit for the innermost main loop only
+			if ($level == $#AWRT_DONE) {
+				Gtk2->main_quit;
+			}
+		},
+	);
+
+	$LAST_PID = $pid;
+
+	# install timeout handler if necessary
+	if ($level == 0) {
+		my $break = 0;
+		my $count = 0;
+		
+		Glib::Timeout->add(10, sub {
+			# pulse progressbars
+			foreach (@ArchWay::Session::WINDOWS) {
+				$_->progressbar->pulse
+					if $count == 0 && $_->can('progressbar');
+
+				$count = ($count + 1) % 3;
+			}
+
+			# check fds
+			Arch::Run::poll(0);
+
+			# clear progressbars after 500ms after last pulse, if no
+			# other tla process has been started by then
+			if ($break) {
+				Glib::Timeout->add(500, sub {
+					if ($LAST_PID == $pid) {
+						foreach (@ArchWay::Session::WINDOWS) {
+							$_->progressbar->set_fraction(0.0)
+								if $_->can('progressbar');
+						}
+					}
+					0;
+				});
+			}
+
+			!$break;
+		});
+
+		Gtk2->main;
+		$break = 1;  # terminate timeout handler
+
+	# simple inner loop
+	} else {
+		Gtk2->main;
+	}
+
+	pop @AWRT_DONE;
+
+	# break next outer main loop if process finished already
+	Gtk2->main_quit
+		if @AWRT_DONE && $AWRT_DONE[-1];
+
+	# does this work?
+	$? = $exit;
+	return wantarray ? split /\015?\012/, $data : $data;
+}
+
 1;
